package net.java.quickcheck;

import static java.lang.String.format;

import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.List;

/**
 * Runner implementation that uses failed test data instances from prior runs.
 * Currently only deserialized test instances will be run. This runner decorator
 * should be used after {@link SerializingRunnerDecorator} serialized failed
 * test data.
 * 
 * Run tests with the system property
 * {@link QuickCheck#SYSTEM_PROPERTY_DESERIALIZE} to enable this decorator.
 */
class DeserializingRunnerDecorator<T> extends AbstractSerializingRunnerDecorator<T> {

	public DeserializingRunnerDecorator(Runner<T> wrapped) {
		super(wrapped);
	}

	@Override
	protected void onForAll() {
		Characteristic<T> characteristic = getDecorated().getCharacteristic();
		List<File> fileNames = getFileNames(characteristic);
		print(format("run deserialized test data (%s found).", fileNames.size()));

		for(int currentRuns = 0; currentRuns < fileNames.size(); currentRuns++) {
			T read = readSerialized(fileNames.get(currentRuns));
			try {
				printCurrentValue(currentRuns + 1, read);
				characteristic.specify(read);
			} catch(Throwable e) {
				throwFailedException(read, e, characteristic, currentRuns + 1);
			}
		}
	}

	private T readSerialized(File fileName) {
		ObjectInputStream objectInputStream = null;
		try {
			objectInputStream = new ObjectInputStream(new FileInputStream(fileName));
			Object readObject = ((SerializedInstanceHolder) objectInputStream.readObject()).getInstance();
			return castToT(readObject);
		} catch(Exception e) {
			throw new RuntimeException(e.getMessage(), e);
		} finally {
			closeQuietly(objectInputStream);
		}
	}

	@SuppressWarnings("unchecked")
	private T castToT(Object readObject) {
		return (T) readObject;
	}

	public List<File> getFileNames(Characteristic<?> characteristic) {
		return discoverFiles(characteristic, new FileDiscovery<List<File>>() {

			private List<File> result = new ArrayList<File>();

			public List<File> getResult() {
				return result;
			}

			public boolean goOn(File file) {
				if(file.exists()) {
					result.add(file);
					return true;
				}
				return false;
			}
		});
	}
}