/*
 *  Licensed to the author under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package net.java.quickcheck;

import static java.lang.String.format;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

/**
 * Runner instance that serializes failed test data instances. To enabled this
 * runner use the system property {@link QuickCheck#SYSTEM_PROPERTY_SERIALIZE}.
 */
class SerializingRunnerDecorator<T> extends
		AbstractSerializingRunnerDecorator<T> {

	private boolean deleteOldFiles;

	public SerializingRunnerDecorator(Runner<T> decorated, boolean deleteOld) {
		super(decorated);
		this.deleteOldFiles = deleteOld;
	}

	@Override
	protected void onForAll() {
		getWriter().println(format("Delete old files %s", deleteOldFiles));
		try {
			if (deleteOldFiles) {
				deleteOldFilesInDir();
			}
			getDecorated().forAll();
		} catch (CharacteristicException e) {
			serialize(e);
			throw e;
		}
	}

	private void deleteOldFilesInDir() {
		File dirForCharacteristic = new File(
				getDirectoryForCharacteristic(getDecorated()
						.getCharacteristic()));
		if (dirForCharacteristic.isDirectory()) {
			for (String file : dirForCharacteristic.list()) {
				new File(dirForCharacteristic, file).delete();
			}
		}
	}

	boolean deleteOldFiles() {
		return deleteOldFiles;
	}

	void serialize(CharacteristicException e) {

		File file = getFileName(e.getCharateristic());
		try {
			writeHolder(file, new SerializableInstanceHolder(e.getInstance()));
		} catch (IOException firstTry) {
			try {
				writeHolder(file, new UnserializableInstanceHolder(e
						.getInstance().toString(), e));
			} catch (IOException holderTry) {
				// TODO test this
				throw new RuntimeException(holderTry);
			}
		}
	}

	private void writeHolder(File file,
			SerializedInstanceHolder serializedInstanceHolderImpl)
			throws IOException {
		ObjectOutputStream objectOutputStream = null;
		try {

			file.getParentFile().mkdirs();
			objectOutputStream = new ObjectOutputStream(new FileOutputStream(
					file));
			objectOutputStream.writeObject(serializedInstanceHolderImpl);
		} finally {
			closeQuietly(objectOutputStream);
		}
	}

	File getFileName(Characteristic<?> c) {
		return discoverFiles(c, new FileDiscovery<File>() {
			private File result;

			public boolean goOn(File file) {
				result = file;
				return file.exists();
			}

			public File getResult() {
				return result;
			}
		});
	}
}
