/*
 *  Licensed to the author under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package net.java.quickcheck.characteristic;

import static java.lang.String.format;

import java.util.ArrayList;
import java.util.List;

/**
 * State expectations about classification frequencies.
 */
public class ClassificationExpectation {

	private final List<Expectation> expectations = new ArrayList<Expectation>();

	/**
	 * Add an expectation about the frequency of a classification value.
	 * <p>
	 * The expression "frequency operator expected" has to be true. For example
	 * "frequency < 10".
	 * 
	 * @param classification
	 *            classification name
	 * @param operator
	 *            operator to check with
	 * @param expected
	 *            expected value
	 * @return this instance
	 */
	public ClassificationExpectation expect(Object classification,
			Operator operator, double expected) {
		expectations.add(new Expectation(classification, operator, expected));
		return this;
	}

	/**
	 * Check that the classification satisfies the expectation.
	 * 
	 * @param classification
	 *            classification to check
	 * @return true if the expectation is met
	 */
	public boolean validate(Classification classification) {
		for (Expectation expectation : expectations) {
			double frequency = classification
					.getFrequency(expectation.classification);
			if (!expectation.check(frequency)) {
				return false;
			}
		}
		return true;
	}

	private static class Expectation {

		private final Object classification;
		private final Operator operator;
		private final double expected;

		public Expectation(Object classification, Operator operator,
				double expected) {
			this.classification = classification;
			this.operator = operator;
			this.expected = expected;

		}

		public boolean check(double actual) {
			return operator.check(actual, expected);
		}

		@Override
		public String toString() {
			return format(
					"Expectation[classification=%s, operator=%s, expected=%s]",
					classification, operator, expected);
		}

	}

	@Override
	public String toString() {
		return format("ClassificationExpectation%s", expectations);
	}

	private static ThreadLocal<Holder> gathered = new ThreadLocal<Holder>() {
		@Override
		protected Holder initialValue() {
			return new Holder();
		}
	};

	private static class Holder {
		public ClassificationExpectation value;
	}

	/**
	 * Store the expectation for later examination.
	 */
	public static void expect(
			ClassificationExpectation classificationExpectation) {
		getHolder().value = classificationExpectation;
	}

	private static Holder getHolder() {
		return gathered.get();
	}

	/**
	 * Get the stored expectation.
	 * 
	 * @return stored value
	 */
	public static ClassificationExpectation gathered() {
		return getHolder().value;
	}

	/**
	 * Clear the stored value.
	 */
	public static void clear() {
		getHolder().value = null;
	}

	static boolean gather() {
		return gathered() != null;
	}
}
