package net.java.quickcheck.junit.support;

import static java.lang.Class.forName;
import static java.util.Collections.singletonList;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.java.quickcheck.Generator;
import net.java.quickcheck.collection.Pair;
import net.java.quickcheck.junit.ForAll;

public class Discovery {

	private static final String DOT_STRING = ".";

	private static List<Method> getMethods(Class<?> clazz, Predicate predicate) {

		List<Method> characteristicMethods = new ArrayList<Method>();
		Method[] declaredMethods = clazz.getDeclaredMethods();
		for (Method method : declaredMethods) {

			if (predicate.evaluate(method)) {
				characteristicMethods.add(method);
			}
		}
		Class<?> superclass = clazz.getSuperclass();
		if (superclass != null) {
			characteristicMethods.addAll(getMethods(superclass, predicate));
		}
		return characteristicMethods;
	}

	static List<Method> discoverGeneratorMethods(Class<?> clazz) {
		return getMethods(clazz, new Predicate() {

			public boolean evaluate(Method method) {
				return Generator.class.isAssignableFrom(method.getReturnType())
						&& method.getParameterTypes().length == 0;
			}
		});
	}

	public static List<Method> getGeneratorMethods(Class<?> testClazz,
			String[] generatorNames) {
		List<Method> matchingMethods = new ArrayList<Method>();
		for (String generatorName : generatorNames) {
			getGeneratorMethods(matchingMethods, testClazz, generatorName);
		}

		if (matchingMethods.size() > 0) {
			return matchingMethods;
		}

		if (discoverGeneratorMethods(testClazz).size() == 1
				&& generatorNames.length == 1
				&& ForAll.NO_GENERATOR.equals(generatorNames[0])) {
			return singletonList(discoverGeneratorMethods(testClazz).get(0));
		}

		throw new IllegalArgumentException(String.format(
				"generator methods (%s) not found ", Arrays
						.toString(generatorNames)));
	}

	private static void getGeneratorMethods(List<Method> matchingMethods,
			Class<?> testClazz, String generatorName) {
		Pair<Class<?>, String> classAndMethod = discoverForGeneratorMethod(
				testClazz, generatorName);
		for (Method method : discoverGeneratorMethods(classAndMethod.getFirst())) {
			if (method.getName().equals(classAndMethod.getSecond())) {
				matchingMethods.add(method);
			}
		}
	}

	private static Pair<Class<?>, String> discoverForGeneratorMethod(
			Class<?> testClazz, String generatorName) {
		if (isStaticGeneratorMethod(generatorName)) {
			int lastDot = generatorName.lastIndexOf(DOT_STRING);
			String usedMethod = generatorName.substring(lastDot + 1);
			String className = generatorName.substring(0, lastDot);
			return new Pair<Class<?>, String>(loadClass(className), usedMethod);
		} 
		return new Pair<Class<?>, String>(testClazz, generatorName);
	}

	private static Class<?> loadClass(String className) {
		try {
			return forName(className);
		} catch (ClassNotFoundException e) {
			throw new RuntimeException(e);
		}
	}

	private static boolean isStaticGeneratorMethod(String generatorName) {
		return generatorName.contains(DOT_STRING);
	}

	public static List<Method> getCharacteristicMethods(Class<?> testClass) {
		return getMethods(testClass, new Predicate() {

			public boolean evaluate(Method method) {
				return method.isAnnotationPresent(ForAll.class);
			}
		});
	}

	public static boolean isGeneratorMethodAnnotation(ForAll annotation) {
		return Arrays.equals(new Class[] { ForAll.NoGenerator.class },
				annotation.generatorClass());
	}

	// TODO move
	private interface Predicate {
		boolean evaluate(Method method);
	}
}
